@@ -15,7 +15,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111-1301  USA
 
-    Copyright Matthias Jentsch, Vitaly Valtman, Christopher Althaus, Markus Endres 2014-2015
+    Copyright Matthias Jentsch, Vitaly Valtman, Ben Wiederhake, Christopher Althaus 2014-2015
 */
 
 #define _GNU_SOURCE
@@ -473,8 +473,8 @@ void write_secret_chat_gw (struct tgl_state *TLS, void *extra, int success, stru
   write_secret_chat_file (TLS);
 }
 
-void tgp_create_group_chat_by_usernames (struct tgl_state *TLS, const char *title, const char **users, int num_users,
-                                         int use_print_names) {
+void tgp_create_group_chat_by_usernames (struct tgl_state *TLS, const char *title, const char **users,
+      int num_users, int use_print_names) {
   tgl_peer_id_t ids[num_users + 1];
   int i, j = 0;
   ids[j++] = TLS->our_id;
@@ -495,10 +495,10 @@ void tgp_create_group_chat_by_usernames (struct tgl_state *TLS, const char *titl
   }
   if (j > 1) {
     tgl_do_create_group_chat (TLS, j, ids, title, (int) strlen(title),
-                              tgp_notify_on_error_gw, g_strdup (title));
+        tgp_notify_on_error_gw, g_strdup (title));
   } else {
-    purple_notify_message (_telegram_protocol, PURPLE_NOTIFY_MSG_INFO, _("Couldn't create group"), _("Please select at least one other user."), NULL,
-                           NULL, NULL);
+    purple_notify_message (_telegram_protocol, PURPLE_NOTIFY_MSG_INFO, _("Couldn't create group"),
+        _("Please select at least one other user."), NULL, NULL, NULL);
   }
 }
 
